@@ -55,7 +55,7 @@ psyco_curs_close(cursorObject *self)
         goto exit;
     }
 
-    if (self->name != NULL) {
+    if (self->qname != NULL) {
         char buffer[128];
         PGTransactionStatusType status;
 
@@ -68,7 +68,7 @@ psyco_curs_close(cursorObject *self)
 
         if (!(status == PQTRANS_UNKNOWN || status == PQTRANS_INERROR)) {
             EXC_IF_NO_MARK(self);
-            PyOS_snprintf(buffer, 127, "CLOSE \"%s\"", self->name);
+            PyOS_snprintf(buffer, 127, "CLOSE %s", self->qname);
             if (pq_execute(self, buffer, 0, 0, 1) == -1) return NULL;
         }
         else {
@@ -267,42 +267,6 @@ _mogrify(PyObject *var, PyObject *fmt, cursorObject *curs, PyObject **new)
     return 0;
 }
 
-static PyObject *_psyco_curs_validate_sql_basic(
-    cursorObject *self, PyObject *sql
-  )
-{
-    /* Performs very basic validation on an incoming SQL string.
-       Returns a new reference to a str instance on success; NULL on failure,
-       after having set an exception. */
-
-    if (!sql || !PyObject_IsTrue(sql)) {
-        psyco_set_error(ProgrammingError, self,
-                         "can't execute an empty query");
-        goto fail;
-    }
-
-    if (Bytes_Check(sql)) {
-        /* Necessary for ref-count symmetry with the unicode case: */
-        Py_INCREF(sql);
-    }
-    else if (PyUnicode_Check(sql)) {
-        char *enc = self->conn->codec;
-        sql = PyUnicode_AsEncodedString(sql, enc, NULL);
-        /* if there was an error during the encoding from unicode to the
-           target encoding, we just let the exception propagate */
-        if (sql == NULL) { goto fail; }
-    }
-    else {
-        /* the  is not unicode or string, raise an error */
-        PyErr_SetString(PyExc_TypeError,
-                        "argument 1 must be a string or unicode object");
-        goto fail;
-    }
-
-    return sql; /* new reference */
-    fail:
-        return NULL;
-}
 
 /* Merge together a query string and its arguments.
  *
@@ -382,7 +346,7 @@ _psyco_curs_execute(cursorObject *self,
     PyObject *fquery, *cvt = NULL;
     const char *scroll;
 
-    operation = _psyco_curs_validate_sql_basic(self, operation);
+    operation = psyco_curs_validate_sql_basic(self, operation);
 
     /* Any failure from here forward should 'goto fail' rather than 'return 0'
        directly. */
@@ -422,10 +386,10 @@ _psyco_curs_execute(cursorObject *self,
             goto exit;
         }
 
-        if (self->name != NULL) {
+        if (self->qname != NULL) {
             self->query = Bytes_FromFormat(
-                "DECLARE \"%s\" %sCURSOR %s HOLD FOR %s",
-                self->name,
+                "DECLARE %s %sCURSOR %s HOLD FOR %s",
+                self->qname,
                 scroll,
                 self->withhold ? "WITH" : "WITHOUT",
                 Bytes_AS_STRING(fquery));
@@ -436,10 +400,10 @@ _psyco_curs_execute(cursorObject *self,
         }
     }
     else {
-        if (self->name != NULL) {
+        if (self->qname != NULL) {
             self->query = Bytes_FromFormat(
-                "DECLARE \"%s\" %sCURSOR %s HOLD FOR %s",
-                self->name,
+                "DECLARE %s %sCURSOR %s HOLD FOR %s",
+                self->qname,
                 scroll,
                 self->withhold ? "WITH" : "WITHOUT",
                 Bytes_AS_STRING(operation));
@@ -579,7 +543,7 @@ _psyco_curs_mogrify(cursorObject *self,
 {
     PyObject *fquery = NULL, *cvt = NULL;
 
-    operation = _psyco_curs_validate_sql_basic(self, operation);
+    operation = psyco_curs_validate_sql_basic(self, operation);
     if (operation == NULL) { goto cleanup; }
 
     Dprintf("psyco_curs_mogrify: starting mogrify");
@@ -628,8 +592,6 @@ psyco_curs_mogrify(cursorObject *self, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
-    EXC_IF_CURS_CLOSED(self);
-
     return _psyco_curs_mogrify(self, operation, vars);
 }
 
@@ -768,13 +730,13 @@ psyco_curs_fetchone(cursorObject *self)
     if (_psyco_curs_prefetch(self) < 0) return NULL;
     EXC_IF_NO_TUPLES(self);
 
-    if (self->name != NULL) {
+    if (self->qname != NULL) {
         char buffer[128];
 
         EXC_IF_NO_MARK(self);
         EXC_IF_ASYNC_IN_PROGRESS(self, fetchone);
         EXC_IF_TPC_PREPARED(self->conn, fetchone);
-        PyOS_snprintf(buffer, 127, "FETCH FORWARD 1 FROM \"%s\"", self->name);
+        PyOS_snprintf(buffer, 127, "FETCH FORWARD 1 FROM %s", self->qname);
         if (pq_execute(self, buffer, 0, 0, self->withhold) == -1) return NULL;
         if (_psyco_curs_prefetch(self) < 0) return NULL;
     }
@@ -823,8 +785,8 @@ psyco_curs_next_named(cursorObject *self)
     if (self->row >= self->rowcount) {
         char buffer[128];
 
-        PyOS_snprintf(buffer, 127, "FETCH FORWARD %ld FROM \"%s\"",
-            self->itersize, self->name);
+        PyOS_snprintf(buffer, 127, "FETCH FORWARD %ld FROM %s",
+            self->itersize, self->qname);
         if (pq_execute(self, buffer, 0, 0, self->withhold) == -1) return NULL;
         if (_psyco_curs_prefetch(self) < 0) return NULL;
     }
@@ -886,14 +848,14 @@ psyco_curs_fetchmany(cursorObject *self, PyObject *args, PyObject *kwords)
     if (_psyco_curs_prefetch(self) < 0) return NULL;
     EXC_IF_NO_TUPLES(self);
 
-    if (self->name != NULL) {
+    if (self->qname != NULL) {
         char buffer[128];
 
         EXC_IF_NO_MARK(self);
         EXC_IF_ASYNC_IN_PROGRESS(self, fetchmany);
         EXC_IF_TPC_PREPARED(self->conn, fetchone);
-        PyOS_snprintf(buffer, 127, "FETCH FORWARD %d FROM \"%s\"",
-            (int)size, self->name);
+        PyOS_snprintf(buffer, 127, "FETCH FORWARD %d FROM %s",
+            (int)size, self->qname);
         if (pq_execute(self, buffer, 0, 0, self->withhold) == -1) { goto exit; }
         if (_psyco_curs_prefetch(self) < 0) { goto exit; }
     }
@@ -962,13 +924,13 @@ psyco_curs_fetchall(cursorObject *self)
     if (_psyco_curs_prefetch(self) < 0) return NULL;
     EXC_IF_NO_TUPLES(self);
 
-    if (self->name != NULL) {
+    if (self->qname != NULL) {
         char buffer[128];
 
         EXC_IF_NO_MARK(self);
         EXC_IF_ASYNC_IN_PROGRESS(self, fetchall);
         EXC_IF_TPC_PREPARED(self->conn, fetchall);
-        PyOS_snprintf(buffer, 127, "FETCH FORWARD ALL FROM \"%s\"", self->name);
+        PyOS_snprintf(buffer, 127, "FETCH FORWARD ALL FROM %s", self->qname);
         if (pq_execute(self, buffer, 0, 0, self->withhold) == -1) { goto exit; }
         if (_psyco_curs_prefetch(self) < 0) { goto exit; }
     }
@@ -1025,10 +987,17 @@ psyco_curs_callproc(cursorObject *self, PyObject *args)
     PyObject *operation = NULL;
     PyObject *res = NULL;
 
-    if (!PyArg_ParseTuple(args, "s#|O",
-          &procname, &procname_len, &parameters
-       ))
-    { goto exit; }
+    int using_dict;
+    PyObject *pname = NULL;
+    PyObject *pnames = NULL;
+    PyObject *pvals = NULL;
+    char *cpname = NULL;
+    char **scpnames = NULL;
+
+    if (!PyArg_ParseTuple(args, "s#|O", &procname, &procname_len,
+                &parameters)) {
+        goto exit;
+    }
 
     EXC_IF_CURS_CLOSED(self);
     EXC_IF_ASYNC_IN_PROGRESS(self, callproc);
@@ -1036,7 +1005,7 @@ psyco_curs_callproc(cursorObject *self, PyObject *args)
 
     if (self->name != NULL) {
         psyco_set_error(ProgrammingError, self,
-                         "can't call .callproc() on named cursors");
+                "can't call .callproc() on named cursors");
         goto exit;
     }
 
@@ -1044,31 +1013,108 @@ psyco_curs_callproc(cursorObject *self, PyObject *args)
         if (-1 == (nparameters = PyObject_Length(parameters))) { goto exit; }
     }
 
-    /* allocate some memory, build the SQL and create a PyString from it */
-    sl = procname_len + 17 + nparameters*3 - (nparameters ? 1 : 0);
-    sql = (char*)PyMem_Malloc(sl);
-    if (sql == NULL) {
-        PyErr_NoMemory();
+    using_dict = nparameters > 0 && PyDict_Check(parameters);
+
+    /* a Dict is complicated; the parameter names go into the query */
+    if (using_dict) {
+        if (!(pnames = PyDict_Keys(parameters))) { goto exit; }
+        if (!(pvals = PyDict_Values(parameters))) { goto exit; }
+
+        sl = procname_len + 17 + nparameters * 5 - (nparameters ? 1 : 0);
+
+        if (!(scpnames = PyMem_New(char *, nparameters))) {
+            PyErr_NoMemory();
+            goto exit;
+        }
+
+        memset(scpnames, 0, sizeof(char *) * nparameters);
+
+        /* each parameter has to be processed; it's a few steps. */
+        for (i = 0; i < nparameters; i++) {
+            /* all errors are RuntimeErrors as they should never occur */
+
+            if (!(pname = PyList_GetItem(pnames, i))) { goto exit; }
+            Py_INCREF(pname);   /* was borrowed */
+
+            /* this also makes a check for keys being strings */
+            if (!(pname = psycopg_ensure_bytes(pname))) { goto exit; }
+            if (!(cpname = Bytes_AsString(pname))) { goto exit; }
+
+            if (!(scpnames[i] = psycopg_escape_identifier(
+                    self->conn, cpname, -1))) {
+                Py_CLEAR(pname);
+                goto exit;
+            }
+
+            Py_CLEAR(pname);
+
+            sl += strlen(scpnames[i]);
+        }
+
+        if (!(sql = (char*)PyMem_Malloc(sl))) {
+            PyErr_NoMemory();
+            goto exit;
+        }
+
+        sprintf(sql, "SELECT * FROM %s(", procname);
+        for (i = 0; i < nparameters; i++) {
+            strcat(sql, scpnames[i]);
+            strcat(sql, ":=%s,");
+        }
+        sql[sl-2] = ')';
+        sql[sl-1] = '\0';
+    }
+
+    /* a list (or None, or empty data structure) is a little bit simpler */
+    else {
+        Py_INCREF(parameters);
+        pvals = parameters;
+
+        sl = procname_len + 17 + nparameters * 3 - (nparameters ? 1 : 0);
+
+        sql = (char*)PyMem_Malloc(sl);
+        if (sql == NULL) {
+            PyErr_NoMemory();
+            goto exit;
+        }
+
+        sprintf(sql, "SELECT * FROM %s(", procname);
+        for (i = 0; i < nparameters; i++) {
+            strcat(sql, "%s,");
+        }
+        sql[sl-2] = ')';
+        sql[sl-1] = '\0';
+    }
+
+    if (!(operation = Bytes_FromString(sql))) {
         goto exit;
     }
 
-    sprintf(sql, "SELECT * FROM %s(", procname);
-    for(i=0; i<nparameters; i++) {
-        strcat(sql, "%s,");
-    }
-    sql[sl-2] = ')';
-    sql[sl-1] = '\0';
-
-    if (!(operation = Bytes_FromString(sql))) { goto exit; }
-
-    if (0 <= _psyco_curs_execute(self, operation, parameters,
-            self->conn->async, 0)) {
-        Py_INCREF(parameters);
-        res = parameters;
+    if (0 <= _psyco_curs_execute(
+            self, operation, pvals, self->conn->async, 0)) {
+        /* The dict case is outside DBAPI scope anyway, so simply return None */
+        if (using_dict) {
+            res = Py_None;
+        }
+        else {
+            res = pvals;
+        }
+        Py_INCREF(res);
     }
 
 exit:
+    if (scpnames != NULL) {
+        for (i = 0; i < nparameters; i++) {
+            if (scpnames[i] != NULL) {
+                PQfreemem(scpnames[i]);
+            }
+        }
+    }
+    PyMem_Del(scpnames);
+    Py_XDECREF(pname);
+    Py_XDECREF(pnames);
     Py_XDECREF(operation);
+    Py_XDECREF(pvals);
     PyMem_Free((void*)sql);
     return res;
 }
@@ -1153,7 +1199,7 @@ psyco_curs_scroll(cursorObject *self, PyObject *args, PyObject *kwargs)
     /* if the cursor is not named we have the full result set and we can do
        our own calculations to scroll; else we just delegate the scrolling
        to the MOVE SQL statement */
-    if (self->name == NULL) {
+    if (self->qname == NULL) {
         if (strcmp(mode, "relative") == 0) {
             newpos = self->row + value;
         } else if (strcmp( mode, "absolute") == 0) {
@@ -1181,11 +1227,11 @@ psyco_curs_scroll(cursorObject *self, PyObject *args, PyObject *kwargs)
         EXC_IF_TPC_PREPARED(self->conn, scroll);
 
         if (strcmp(mode, "absolute") == 0) {
-            PyOS_snprintf(buffer, 127, "MOVE ABSOLUTE %d FROM \"%s\"",
-                value, self->name);
+            PyOS_snprintf(buffer, 127, "MOVE ABSOLUTE %d FROM %s",
+                value, self->qname);
         }
         else {
-            PyOS_snprintf(buffer, 127, "MOVE %d FROM \"%s\"", value, self->name);
+            PyOS_snprintf(buffer, 127, "MOVE %d FROM %s", value, self->qname);
         }
         if (pq_execute(self, buffer, 0, 0, self->withhold) == -1) return NULL;
         if (_psyco_curs_prefetch(self) < 0) return NULL;
@@ -1373,12 +1419,12 @@ psyco_curs_copy_from(cursorObject *self, PyObject *args, PyObject *kwargs)
         goto exit;
 
     if (!(quoted_delimiter = psycopg_escape_string(
-            self->conn, sep, 0, NULL, NULL))) {
+            self->conn, sep, -1, NULL, NULL))) {
         goto exit;
     }
 
     if (!(quoted_null = psycopg_escape_string(
-            self->conn, null, 0, NULL, NULL))) {
+            self->conn, null, -1, NULL, NULL))) {
         goto exit;
     }
 
@@ -1467,12 +1513,12 @@ psyco_curs_copy_to(cursorObject *self, PyObject *args, PyObject *kwargs)
         goto exit;
 
     if (!(quoted_delimiter = psycopg_escape_string(
-            self->conn, sep, 0, NULL, NULL))) {
+            self->conn, sep, -1, NULL, NULL))) {
         goto exit;
     }
 
     if (!(quoted_null = psycopg_escape_string(
-            self->conn, null, 0, NULL, NULL))) {
+            self->conn, null, -1, NULL, NULL))) {
         goto exit;
     }
 
@@ -1538,7 +1584,7 @@ psyco_curs_copy_expert(cursorObject *self, PyObject *args, PyObject *kwargs)
     EXC_IF_GREEN(copy_expert);
     EXC_IF_TPC_PREPARED(self->conn, copy_expert);
 
-    sql = _psyco_curs_validate_sql_basic(self, sql);
+    sql = psyco_curs_validate_sql_basic(self, sql);
 
     /* Any failure from here forward should 'goto exit' rather than
        'return NULL' directly. */
@@ -1815,7 +1861,10 @@ cursor_setup(cursorObject *self, connectionObject *conn, const char *name)
     Dprintf("cursor_setup: parameters: name = %s, conn = %p", name, conn);
 
     if (name) {
-        if (!(self->name = psycopg_escape_identifier_easy(name, 0))) {
+        if (0 > psycopg_strdup(&self->name, name, -1)) {
+            return -1;
+        }
+        if (!(self->qname = psycopg_escape_identifier(conn, name, -1))) {
             return -1;
         }
     }
@@ -1891,6 +1940,7 @@ cursor_dealloc(PyObject* obj)
     cursor_clear(self);
 
     PyMem_Free(self->name);
+    PQfreemem(self->qname);
 
     CLEARPGRES(self->pgres);
 
