@@ -86,9 +86,7 @@ psyco_lobj_write(lobjectObject *self, PyObject *args)
         data = obj;
     }
     else if (PyUnicode_Check(obj)) {
-        if (!(data = PyUnicode_AsEncodedString(obj, self->conn->codec, NULL))) {
-            goto exit;
-        }
+        if (!(data = conn_encode(self->conn, obj))) { goto exit; }
     }
     else {
         PyErr_Format(PyExc_TypeError,
@@ -150,7 +148,7 @@ psyco_lobj_read(lobjectObject *self, PyObject *args)
     if (self->mode & LOBJECT_BINARY) {
         res = Bytes_FromStringAndSize(buffer, size);
     } else {
-        res = PyUnicode_Decode(buffer, size, self->conn->codec, NULL);
+        res = conn_decode(self->conn, buffer, size);
     }
     PyMem_Free(buffer);
 
@@ -266,8 +264,6 @@ psyco_lobj_get_closed(lobjectObject *self, void *closure)
     return closed;
 }
 
-#if PG_VERSION_HEX >= 0x080300
-
 #define psyco_lobj_truncate_doc \
 "truncate(len=0) -- Truncate large object to given size."
 
@@ -327,10 +323,8 @@ static struct PyMethodDef lobjectObject_methods[] = {
      METH_NOARGS, psyco_lobj_unlink_doc},
     {"export",(PyCFunction)psyco_lobj_export,
      METH_VARARGS, psyco_lobj_export_doc},
-#if PG_VERSION_HEX >= 0x080300
     {"truncate",(PyCFunction)psyco_lobj_truncate,
      METH_VARARGS, psyco_lobj_truncate_doc},
-#endif /* PG_VERSION_HEX >= 0x080300 */
     {NULL}
 };
 
@@ -404,7 +398,7 @@ static int
 lobject_init(PyObject *obj, PyObject *args, PyObject *kwds)
 {
     Oid oid = InvalidOid, new_oid = InvalidOid;
-    const char *smode = "";
+    const char *smode = NULL;
     const char *new_file = NULL;
     PyObject *conn = NULL;
 
@@ -413,6 +407,9 @@ lobject_init(PyObject *obj, PyObject *args, PyObject *kwds)
          &oid, &smode, &new_oid, &new_file))
         return -1;
 
+    if (!smode)
+        smode = "";
+
     return lobject_setup((lobjectObject *)obj,
         (connectionObject *)conn, oid, smode, new_oid, new_file);
 }
@@ -475,6 +472,3 @@ PyTypeObject lobjectType = {
     0,          /*tp_alloc*/
     lobject_new, /*tp_new*/
 };
-
-#endif
-
