@@ -153,15 +153,6 @@ microprotocols_adapt(PyObject *obj, PyObject *proto, PyObject *alt)
         return adapted;
     }
 
-    /* Check if a superclass can be adapted and use the same adapter. */
-    if (!(adapter = _get_superclass_adapter(obj, proto))) {
-        return NULL;
-    }
-    if (Py_None != adapter) {
-        adapted = PyObject_CallFunctionObjArgs(adapter, obj, NULL);
-        return adapted;
-    }
-
     /* try to have the protocol adapt this object*/
     if ((meth = PyObject_GetAttrString(proto, "__adapt__"))) {
         adapted = PyObject_CallFunctionObjArgs(meth, obj, NULL);
@@ -181,7 +172,7 @@ microprotocols_adapt(PyObject *obj, PyObject *proto, PyObject *alt)
         PyErr_Clear();
     }
 
-    /* and finally try to have the object adapt itself */
+    /* then try to have the object adapt itself */
     if ((meth = PyObject_GetAttrString(obj, "__conform__"))) {
         adapted = PyObject_CallFunctionObjArgs(meth, proto, NULL);
         Py_DECREF(meth);
@@ -200,10 +191,19 @@ microprotocols_adapt(PyObject *obj, PyObject *proto, PyObject *alt)
         PyErr_Clear();
     }
 
+    /* Finally check if a superclass can be adapted and use the same adapter. */
+    if (!(adapter = _get_superclass_adapter(obj, proto))) {
+        return NULL;
+    }
+    if (Py_None != adapter) {
+        adapted = PyObject_CallFunctionObjArgs(adapter, obj, NULL);
+        return adapted;
+    }
+
     /* else set the right exception and return NULL */
     PyOS_snprintf(buffer, 255, "can't adapt type '%s'",
         Py_TYPE(obj)->tp_name);
-    psyco_set_error(ProgrammingError, NULL, buffer, NULL, NULL);
+    psyco_set_error(ProgrammingError, NULL, buffer);
     return NULL;
 }
 
@@ -251,9 +251,7 @@ microprotocol_getquoted(PyObject *obj, connectionObject *conn)
     /* Convert to bytes. */
     if (res && PyUnicode_CheckExact(res)) {
         PyObject *b;
-        const char *codec;
-        codec = (conn && conn->codec) ? conn->codec : "utf8";
-        b = PyUnicode_AsEncodedString(res, codec, NULL);
+        b = conn_encode(conn, res);
         Py_DECREF(res);
         res = b;
     }
