@@ -26,6 +26,10 @@
 #ifndef PSYCOPG_H
 #define PSYCOPG_H 1
 
+#if PG_VERSION_NUM < 90100
+#error "Psycopg requires PostgreSQL client library (libpq) >= 9.1"
+#endif
+
 #define PY_SSIZE_T_CLEAN
 #include <Python.h>
 #include <libpq-fe.h>
@@ -117,20 +121,29 @@ HIDDEN PyObject *psyco_GetDecimalType(void);
 /* forward declarations */
 typedef struct cursorObject cursorObject;
 typedef struct connectionObject connectionObject;
+typedef struct replicationMessageObject replicationMessageObject;
 
 /* some utility functions */
 RAISES HIDDEN PyObject *psyco_set_error(PyObject *exc, cursorObject *curs, const char *msg);
 
 HIDDEN char *psycopg_escape_string(connectionObject *conn,
               const char *from, Py_ssize_t len, char *to, Py_ssize_t *tolen);
-HIDDEN char *psycopg_escape_identifier_easy(const char *from, Py_ssize_t len);
+HIDDEN char *psycopg_escape_identifier(connectionObject *conn,
+              const char *str, Py_ssize_t len);
 HIDDEN int psycopg_strdup(char **to, const char *from, Py_ssize_t len);
 HIDDEN int psycopg_is_text_file(PyObject *f);
+HIDDEN PyObject *psycopg_text_from_chars_safe(
+        const char *str, Py_ssize_t len, PyObject *decoder);
 
 STEALS(1) HIDDEN PyObject * psycopg_ensure_bytes(PyObject *obj);
 
 STEALS(1) HIDDEN PyObject * psycopg_ensure_text(PyObject *obj);
 
+HIDDEN PyObject *psycopg_dict_from_conninfo_options(PQconninfoOption *options,
+              int include_password);
+
+HIDDEN PyObject *psycopg_make_dsn(PyObject *dsn, PyObject *kwargs);
+
 /* Exceptions docstrings */
 #define Error_doc \
 "Base class for error exceptions."
