@@ -25,13 +25,17 @@
 
 #include "mxDateTime.h"
 
-int
+
+/* Return 0 on success, -1 on failure, but don't set an exception */
+
+static int
 psyco_typecast_mxdatetime_init(void)
 {
     Dprintf("psyco_typecast_mxdatetime_init: mx.DateTime init");
 
-    if(mxDateTime_ImportModuleAndAPI()) {
-        PyErr_SetString(PyExc_ImportError, "mx.DateTime initialization failed");
+    if (mxDateTime_ImportModuleAndAPI()) {
+        Dprintf("psyco_typecast_mxdatetime_init: mx.DateTime initialization failed");
+        PyErr_Clear();
         return -1;
     }
     return 0;
@@ -47,7 +51,7 @@ typecast_MXDATE_cast(const char *str, Py_ssize_t len, PyObject *curs)
     int hh=0, mm=0, ss=0, us=0, tz=0;
     const char *tp = NULL;
 
-    if (str == NULL) {Py_INCREF(Py_None); return Py_None;}
+    if (str == NULL) { Py_RETURN_NONE; }
 
     Dprintf("typecast_MXDATE_cast: s = %s", str);
 
@@ -95,7 +99,7 @@ typecast_MXTIME_cast(const char *str, Py_ssize_t len, PyObject *curs)
 {
     int n, hh=0, mm=0, ss=0, us=0, tz=0;
 
-    if (str == NULL) {Py_INCREF(Py_None); return Py_None;}
+    if (str == NULL) { Py_RETURN_NONE; }
 
     Dprintf("typecast_MXTIME_cast: s = %s", str);
 
@@ -125,7 +129,7 @@ typecast_MXINTERVAL_cast(const char *str, Py_ssize_t len, PyObject *curs)
     double v = 0.0, sign = 1.0;
     int part = 0;
 
-    if (str == NULL) {Py_INCREF(Py_None); return Py_None;}
+    if (str == NULL) { Py_RETURN_NONE; }
 
     Dprintf("typecast_MXINTERVAL_cast: s = %s", str);
 
@@ -138,7 +142,7 @@ typecast_MXINTERVAL_cast(const char *str, Py_ssize_t len, PyObject *curs)
 
         case '0': case '1': case '2': case '3': case '4':
         case '5': case '6': case '7': case '8': case '9':
-            v = v*10 + (double)*str - (double)'0';
+            v = v * 10.0 + (double)(*str - '0');
             Dprintf("typecast_MXINTERVAL_cast: v = %f", v);
             if (part == 6){
                 denominator *= 10;
@@ -244,5 +248,6 @@ typecast_MXINTERVAL_cast(const char *str, Py_ssize_t len, PyObject *curs)
 #define typecast_TIME_cast typecast_MXTIME_cast
 #define typecast_INTERVAL_cast typecast_MXINTERVAL_cast
 #define typecast_DATETIME_cast typecast_MXDATE_cast
+#define typecast_DATETIMETZ_cast typecast_MXDATE_cast
 #endif
 
