@@ -23,10 +23,15 @@
 # License for more details.
 
 import math
-import unittest
 import psycopg2
-from psycopg2.tz import FixedOffsetTimezone
-from testconfig import dsn
+from psycopg2.tz import FixedOffsetTimezone, ZERO
+from testutils import unittest, ConnectingTestCase, skip_before_postgres
+
+
+def total_seconds(d):
+    """Return total number of seconds of a timedelta as a float."""
+    return d.days * 24 * 60 * 60 + d.seconds + d.microseconds / 1000000.0
+
 
 class CommonDatetimeTestsMixin:
 
@@ -78,7 +83,7 @@ class CommonDatetimeTestsMixin:
         value = self.DATETIME(None, self.curs)
         self.assertEqual(value, None)
 
-    def test_parse_incomplete_time(self):
+    def test_parse_incomplete_datetime(self):
         self.assertRaises(psycopg2.DataError,
                           self.DATETIME, '2007', self.curs)
         self.assertRaises(psycopg2.DataError,
@@ -93,20 +98,17 @@ class CommonDatetimeTestsMixin:
         self.assertEqual(value, None)
 
 
-class DatetimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
+class DatetimeTests(ConnectingTestCase, CommonDatetimeTestsMixin):
     """Tests for the datetime based date handling in psycopg2."""
 
     def setUp(self):
-        self.conn = psycopg2.connect(dsn)
+        ConnectingTestCase.setUp(self)
         self.curs = self.conn.cursor()
         self.DATE = psycopg2.extensions.PYDATE
         self.TIME = psycopg2.extensions.PYTIME
         self.DATETIME = psycopg2.extensions.PYDATETIME
         self.INTERVAL = psycopg2.extensions.PYINTERVAL
 
-    def tearDown(self):
-        self.conn.close()
-
     def test_parse_bc_date(self):
         # datetime does not support BC dates
         self.assertRaises(ValueError, self.DATE, '00042-01-01 BC', self.curs)
@@ -148,10 +150,10 @@ class DatetimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
         # The Python datetime module does not support time zone
         # offsets that are not a whole number of minutes.
         # We round the offset to the nearest minute.
-        self.check_time_tz("+01:15:00",  60 * (60 + 15))
-        self.check_time_tz("+01:15:29",  60 * (60 + 15))
-        self.check_time_tz("+01:15:30",  60 * (60 + 16))
-        self.check_time_tz("+01:15:59",  60 * (60 + 16))
+        self.check_time_tz("+01:15:00", 60 * (60 + 15))
+        self.check_time_tz("+01:15:29", 60 * (60 + 15))
+        self.check_time_tz("+01:15:30", 60 * (60 + 16))
+        self.check_time_tz("+01:15:59", 60 * (60 + 16))
         self.check_time_tz("-01:15:00", -60 * (60 + 15))
         self.check_time_tz("-01:15:29", -60 * (60 + 15))
         self.check_time_tz("-01:15:30", -60 * (60 + 16))
@@ -184,10 +186,10 @@ class DatetimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
         # The Python datetime module does not support time zone
         # offsets that are not a whole number of minutes.
         # We round the offset to the nearest minute.
-        self.check_datetime_tz("+01:15:00",  60 * (60 + 15))
-        self.check_datetime_tz("+01:15:29",  60 * (60 + 15))
-        self.check_datetime_tz("+01:15:30",  60 * (60 + 16))
-        self.check_datetime_tz("+01:15:59",  60 * (60 + 16))
+        self.check_datetime_tz("+01:15:00", 60 * (60 + 15))
+        self.check_datetime_tz("+01:15:29", 60 * (60 + 15))
+        self.check_datetime_tz("+01:15:30", 60 * (60 + 16))
+        self.check_datetime_tz("+01:15:59", 60 * (60 + 16))
         self.check_datetime_tz("-01:15:00", -60 * (60 + 15))
         self.check_datetime_tz("-01:15:29", -60 * (60 + 15))
         self.check_datetime_tz("-01:15:30", -60 * (60 + 16))
@@ -217,6 +219,14 @@ class DatetimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
         self.assertEqual(value.seconds, 41103)
         self.assertEqual(value.microseconds, 876544)
 
+    def test_parse_infinity(self):
+        value = self.DATETIME('-infinity', self.curs)
+        self.assertEqual(str(value), '0001-01-01 00:00:00')
+        value = self.DATETIME('infinity', self.curs)
+        self.assertEqual(str(value), '9999-12-31 23:59:59.999999')
+        value = self.DATE('infinity', self.curs)
+        self.assertEqual(str(value), '9999-12-31')
+
     def test_adapt_date(self):
         from datetime import date
         value = self.execute('select (%s)::date::text',
@@ -244,7 +254,7 @@ class DatetimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
         self.assertEqual(seconds, 3674096)
         self.assertEqual(int(round((value - seconds) * 1000000)), 123456)
 
-    def test_adapt_megative_timedelta(self):
+    def test_adapt_negative_timedelta(self):
         from datetime import timedelta
         value = self.execute('select extract(epoch from (%s)::interval)',
                              [timedelta(days=-42, seconds=45296,
@@ -265,25 +275,35 @@ class DatetimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
 
     def test_type_roundtrip_date(self):
         from datetime import date
-        self._test_type_roundtrip(date(2010,5,3))
+        self._test_type_roundtrip(date(2010, 5, 3))
 
     def test_type_roundtrip_datetime(self):
         from datetime import datetime
-        dt = self._test_type_roundtrip(datetime(2010,5,3,10,20,30))
+        dt = self._test_type_roundtrip(datetime(2010, 5, 3, 10, 20, 30))
         self.assertEqual(None, dt.tzinfo)
 
     def test_type_roundtrip_datetimetz(self):
         from datetime import datetime
         import psycopg2.tz
-        tz = psycopg2.tz.FixedOffsetTimezone(8*60)
-        dt1 = datetime(2010,5,3,10,20,30, tzinfo=tz)
+        tz = psycopg2.tz.FixedOffsetTimezone(8 * 60)
+        dt1 = datetime(2010, 5, 3, 10, 20, 30, tzinfo=tz)
         dt2 = self._test_type_roundtrip(dt1)
         self.assertNotEqual(None, dt2.tzinfo)
         self.assertEqual(dt1, dt2)
 
     def test_type_roundtrip_time(self):
         from datetime import time
-        self._test_type_roundtrip(time(10,20,30))
+        tm = self._test_type_roundtrip(time(10, 20, 30))
+        self.assertEqual(None, tm.tzinfo)
+
+    def test_type_roundtrip_timetz(self):
+        from datetime import time
+        import psycopg2.tz
+        tz = psycopg2.tz.FixedOffsetTimezone(8 * 60)
+        tm1 = time(10, 20, 30, tzinfo=tz)
+        tm2 = self._test_type_roundtrip(tm1)
+        self.assertNotEqual(None, tm2.tzinfo)
+        self.assertEqual(tm1, tm2)
 
     def test_type_roundtrip_interval(self):
         from datetime import timedelta
@@ -291,31 +311,143 @@ class DatetimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
 
     def test_type_roundtrip_date_array(self):
         from datetime import date
-        self._test_type_roundtrip_array(date(2010,5,3))
+        self._test_type_roundtrip_array(date(2010, 5, 3))
 
     def test_type_roundtrip_datetime_array(self):
         from datetime import datetime
-        self._test_type_roundtrip_array(datetime(2010,5,3,10,20,30))
+        self._test_type_roundtrip_array(datetime(2010, 5, 3, 10, 20, 30))
+
+    def test_type_roundtrip_datetimetz_array(self):
+        from datetime import datetime
+        self._test_type_roundtrip_array(
+            datetime(2010, 5, 3, 10, 20, 30, tzinfo=FixedOffsetTimezone(0)))
 
     def test_type_roundtrip_time_array(self):
         from datetime import time
-        self._test_type_roundtrip_array(time(10,20,30))
+        self._test_type_roundtrip_array(time(10, 20, 30))
 
     def test_type_roundtrip_interval_array(self):
         from datetime import timedelta
         self._test_type_roundtrip_array(timedelta(seconds=30))
 
+    @skip_before_postgres(8, 1)
+    def test_time_24(self):
+        from datetime import time
+
+        t = self.execute("select '24:00'::time;")
+        self.assertEqual(t, time(0, 0))
+
+        t = self.execute("select '24:00+05'::timetz;")
+        self.assertEqual(t, time(0, 0, tzinfo=FixedOffsetTimezone(300)))
+
+        t = self.execute("select '24:00+05:30'::timetz;")
+        self.assertEqual(t, time(0, 0, tzinfo=FixedOffsetTimezone(330)))
+
+    @skip_before_postgres(8, 1)
+    def test_large_interval(self):
+        t = self.execute("select '999999:00:00'::interval")
+        self.assertEqual(total_seconds(t), 999999 * 60 * 60)
+
+        t = self.execute("select '-999999:00:00'::interval")
+        self.assertEqual(total_seconds(t), -999999 * 60 * 60)
+
+        t = self.execute("select '999999:00:00.1'::interval")
+        self.assertEqual(total_seconds(t), 999999 * 60 * 60 + 0.1)
+
+        t = self.execute("select '999999:00:00.9'::interval")
+        self.assertEqual(total_seconds(t), 999999 * 60 * 60 + 0.9)
+
+        t = self.execute("select '-999999:00:00.1'::interval")
+        self.assertEqual(total_seconds(t), -999999 * 60 * 60 - 0.1)
+
+        t = self.execute("select '-999999:00:00.9'::interval")
+        self.assertEqual(total_seconds(t), -999999 * 60 * 60 - 0.9)
+
+    def test_micros_rounding(self):
+        t = self.execute("select '0.1'::interval")
+        self.assertEqual(total_seconds(t), 0.1)
+
+        t = self.execute("select '0.01'::interval")
+        self.assertEqual(total_seconds(t), 0.01)
+
+        t = self.execute("select '0.000001'::interval")
+        self.assertEqual(total_seconds(t), 1e-6)
+
+        t = self.execute("select '0.0000004'::interval")
+        self.assertEqual(total_seconds(t), 0)
+
+        t = self.execute("select '0.0000006'::interval")
+        self.assertEqual(total_seconds(t), 1e-6)
+
+    def test_interval_overflow(self):
+        cur = self.conn.cursor()
+        # hack a cursor to receive values too extreme to be represented
+        # but still I want an error, not a random number
+        psycopg2.extensions.register_type(
+            psycopg2.extensions.new_type(
+                psycopg2.STRING.values, 'WAT', psycopg2.extensions.INTERVAL),
+            cur)
+
+        def f(val):
+            cur.execute("select '%s'::text" % val)
+            return cur.fetchone()[0]
+
+        self.assertRaises(OverflowError, f, '100000000000000000:00:00')
+        self.assertRaises(OverflowError, f, '00:100000000000000000:00:00')
+        self.assertRaises(OverflowError, f, '00:00:100000000000000000:00')
+        self.assertRaises(OverflowError, f, '00:00:00.100000000000000000')
+
+    def test_adapt_infinity_tz(self):
+        from datetime import datetime
+
+        t = self.execute("select 'infinity'::timestamp")
+        self.assert_(t.tzinfo is None)
+        self.assert_(t > datetime(4000, 1, 1))
+
+        t = self.execute("select '-infinity'::timestamp")
+        self.assert_(t.tzinfo is None)
+        self.assert_(t < datetime(1000, 1, 1))
+
+        t = self.execute("select 'infinity'::timestamptz")
+        self.assert_(t.tzinfo is not None)
+        self.assert_(t > datetime(4000, 1, 1, tzinfo=FixedOffsetTimezone()))
+
+        t = self.execute("select '-infinity'::timestamptz")
+        self.assert_(t.tzinfo is not None)
+        self.assert_(t < datetime(1000, 1, 1, tzinfo=FixedOffsetTimezone()))
+
+    def test_redshift_day(self):
+        # Redshift is reported returning 1 day interval as microsec (bug #558)
+        cur = self.conn.cursor()
+        psycopg2.extensions.register_type(
+            psycopg2.extensions.new_type(
+                psycopg2.STRING.values, 'WAT', psycopg2.extensions.INTERVAL),
+            cur)
+
+        from datetime import timedelta
+        for s, v in [
+            ('0', timedelta(0)),
+            ('1', timedelta(microseconds=1)),
+            ('-1', timedelta(microseconds=-1)),
+            ('1000000', timedelta(seconds=1)),
+            ('86400000000', timedelta(days=1)),
+            ('-86400000000', timedelta(days=-1)),
+        ]:
+            cur.execute("select %s::text", (s,))
+            r = cur.fetchone()[0]
+            self.assertEqual(r, v, "%s -> %s != %s" % (s, r, v))
+
 
 # Only run the datetime tests if psycopg was compiled with support.
 if not hasattr(psycopg2.extensions, 'PYDATETIME'):
     del DatetimeTests
 
 
-class mxDateTimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
+class mxDateTimeTests(ConnectingTestCase, CommonDatetimeTestsMixin):
     """Tests for the mx.DateTime based date handling in psycopg2."""
 
     def setUp(self):
-        self.conn = psycopg2.connect(dsn)
+        ConnectingTestCase.setUp(self)
         self.curs = self.conn.cursor()
         self.DATE = psycopg2._psycopg.MXDATE
         self.TIME = psycopg2._psycopg.MXTIME
@@ -328,8 +460,10 @@ class mxDateTimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
         psycopg2.extensions.register_type(self.INTERVAL, self.conn)
         psycopg2.extensions.register_type(psycopg2.extensions.MXDATEARRAY, self.conn)
         psycopg2.extensions.register_type(psycopg2.extensions.MXTIMEARRAY, self.conn)
-        psycopg2.extensions.register_type(psycopg2.extensions.MXDATETIMEARRAY, self.conn)
-        psycopg2.extensions.register_type(psycopg2.extensions.MXINTERVALARRAY, self.conn)
+        psycopg2.extensions.register_type(
+            psycopg2.extensions.MXDATETIMEARRAY, self.conn)
+        psycopg2.extensions.register_type(
+            psycopg2.extensions.MXINTERVALARRAY, self.conn)
 
     def tearDown(self):
         self.conn.close()
@@ -432,7 +566,7 @@ class mxDateTimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
         self.assertEqual(seconds, 3674096)
         self.assertEqual(int(round((value - seconds) * 1000000)), 123456)
 
-    def test_adapt_megative_timedelta(self):
+    def test_adapt_negative_timedelta(self):
         from mx.DateTime import DateTimeDeltaFrom
         value = self.execute('select extract(epoch from (%s)::interval)',
                              [DateTimeDeltaFrom(days=-42,
@@ -452,15 +586,15 @@ class mxDateTimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
 
     def test_type_roundtrip_date(self):
         from mx.DateTime import Date
-        self._test_type_roundtrip(Date(2010,5,3))
+        self._test_type_roundtrip(Date(2010, 5, 3))
 
     def test_type_roundtrip_datetime(self):
         from mx.DateTime import DateTime
-        self._test_type_roundtrip(DateTime(2010,5,3,10,20,30))
+        self._test_type_roundtrip(DateTime(2010, 5, 3, 10, 20, 30))
 
     def test_type_roundtrip_time(self):
         from mx.DateTime import Time
-        self._test_type_roundtrip(Time(10,20,30))
+        self._test_type_roundtrip(Time(10, 20, 30))
 
     def test_type_roundtrip_interval(self):
         from mx.DateTime import DateTimeDeltaFrom
@@ -468,15 +602,15 @@ class mxDateTimeTests(unittest.TestCase, CommonDatetimeTestsMixin):
 
     def test_type_roundtrip_date_array(self):
         from mx.DateTime import Date
-        self._test_type_roundtrip_array(Date(2010,5,3))
+        self._test_type_roundtrip_array(Date(2010, 5, 3))
 
     def test_type_roundtrip_datetime_array(self):
         from mx.DateTime import DateTime
-        self._test_type_roundtrip_array(DateTime(2010,5,3,10,20,30))
+        self._test_type_roundtrip_array(DateTime(2010, 5, 3, 10, 20, 30))
 
     def test_type_roundtrip_time_array(self):
         from mx.DateTime import Time
-        self._test_type_roundtrip_array(Time(10,20,30))
+        self._test_type_roundtrip_array(Time(10, 20, 30))
 
     def test_type_roundtrip_interval_array(self):
         from mx.DateTime import DateTimeDeltaFrom
@@ -504,7 +638,8 @@ class FromTicksTestCase(unittest.TestCase):
     def test_date_value_error_sec_59_99(self):
         from datetime import date
         s = psycopg2.DateFromTicks(1273173119.99992)
-        self.assertEqual(s.adapted, date(2010, 5, 6))
+        # The returned date is local
+        self.assert_(s.adapted in [date(2010, 5, 6), date(2010, 5, 7)])
 
     def test_time_value_error_sec_59_99(self):
         from datetime import time
@@ -513,6 +648,59 @@ class FromTicksTestCase(unittest.TestCase):
             time(0, 11, 59, 999920))
 
 
+class FixedOffsetTimezoneTests(unittest.TestCase):
+
+    def test_init_with_no_args(self):
+        tzinfo = FixedOffsetTimezone()
+        self.assert_(tzinfo._offset is ZERO)
+        self.assert_(tzinfo._name is None)
+
+    def test_repr_with_positive_offset(self):
+        tzinfo = FixedOffsetTimezone(5 * 60)
+        self.assertEqual(repr(tzinfo),
+            "psycopg2.tz.FixedOffsetTimezone(offset=300, name=None)")
+
+    def test_repr_with_negative_offset(self):
+        tzinfo = FixedOffsetTimezone(-5 * 60)
+        self.assertEqual(repr(tzinfo),
+            "psycopg2.tz.FixedOffsetTimezone(offset=-300, name=None)")
+
+    def test_repr_with_name(self):
+        tzinfo = FixedOffsetTimezone(name="FOO")
+        self.assertEqual(repr(tzinfo),
+            "psycopg2.tz.FixedOffsetTimezone(offset=0, name='FOO')")
+
+    def test_instance_caching(self):
+        self.assert_(FixedOffsetTimezone(name="FOO")
+            is FixedOffsetTimezone(name="FOO"))
+        self.assert_(FixedOffsetTimezone(7 * 60)
+            is FixedOffsetTimezone(7 * 60))
+        self.assert_(FixedOffsetTimezone(-9 * 60, 'FOO')
+            is FixedOffsetTimezone(-9 * 60, 'FOO'))
+        self.assert_(FixedOffsetTimezone(9 * 60)
+            is not FixedOffsetTimezone(9 * 60, 'FOO'))
+        self.assert_(FixedOffsetTimezone(name='FOO')
+            is not FixedOffsetTimezone(9 * 60, 'FOO'))
+
+    def test_pickle(self):
+        # ticket #135
+        import pickle
+
+        tz11 = FixedOffsetTimezone(60)
+        tz12 = FixedOffsetTimezone(120)
+        for proto in [-1, 0, 1, 2]:
+            tz21, tz22 = pickle.loads(pickle.dumps([tz11, tz12], proto))
+            self.assertEqual(tz11, tz21)
+            self.assertEqual(tz12, tz22)
+
+        tz11 = FixedOffsetTimezone(60, name='foo')
+        tz12 = FixedOffsetTimezone(120, name='bar')
+        for proto in [-1, 0, 1, 2]:
+            tz21, tz22 = pickle.loads(pickle.dumps([tz11, tz12], proto))
+            self.assertEqual(tz11, tz21)
+            self.assertEqual(tz12, tz22)
+
+
 def test_suite():
     return unittest.TestLoader().loadTestsFromName(__name__)
 
