@@ -22,19 +22,25 @@
 # FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 # License for more details.
 
-from testutils import unittest, skip_before_python, skip_before_postgres
-from testutils import ConnectingTestCase, skip_copy_if_green
+import os
+import sys
+from subprocess import Popen
+
+from testutils import (unittest, skip_before_python, skip_before_postgres,
+    ConnectingTestCase, skip_copy_if_green, script_to_py3, slow)
 
 import psycopg2
 
+
 class ConnectTestCase(unittest.TestCase):
     def setUp(self):
         self.args = None
-        def conect_stub(dsn, connection_factory=None, async=False):
-            self.args = (dsn, connection_factory, async)
+
+        def connect_stub(dsn, connection_factory=None, async_=False):
+            self.args = (dsn, connection_factory, async_)
 
         self._connect_orig = psycopg2._connect
-        psycopg2._connect = conect_stub
+        psycopg2._connect = connect_stub
 
     def tearDown(self):
         psycopg2._connect = self._connect_orig
@@ -42,9 +48,9 @@ class ConnectTestCase(unittest.TestCase):
     def test_there_has_to_be_something(self):
         self.assertRaises(TypeError, psycopg2.connect)
         self.assertRaises(TypeError, psycopg2.connect,
-            connection_factory=lambda dsn, async=False: None)
+            connection_factory=lambda dsn, async_=False: None)
         self.assertRaises(TypeError, psycopg2.connect,
-            async=True)
+            async_=True)
 
     def test_no_keywords(self):
         psycopg2.connect('')
@@ -53,8 +59,8 @@ class ConnectTestCase(unittest.TestCase):
         self.assertEqual(self.args[2], False)
 
     def test_dsn(self):
-        psycopg2.connect('dbname=blah x=y')
-        self.assertEqual(self.args[0], 'dbname=blah x=y')
+        psycopg2.connect('dbname=blah host=y')
+        self.assertEqual(self.args[0], 'dbname=blah host=y')
         self.assertEqual(self.args[1], None)
         self.assertEqual(self.args[2], False)
 
@@ -79,37 +85,43 @@ class ConnectTestCase(unittest.TestCase):
         self.assertEqual(len(self.args[0].split()), 4)
 
     def test_generic_keywords(self):
-        psycopg2.connect(foo='bar')
-        self.assertEqual(self.args[0], 'foo=bar')
+        psycopg2.connect(options='stuff')
+        self.assertEqual(self.args[0], 'options=stuff')
 
     def test_factory(self):
-        def f(dsn, async=False):
+        def f(dsn, async_=False):
             pass
 
-        psycopg2.connect(database='foo', bar='baz', connection_factory=f)
-        self.assertEqual(self.args[0], 'dbname=foo bar=baz')
+        psycopg2.connect(database='foo', host='baz', connection_factory=f)
+        self.assertDsnEqual(self.args[0], 'dbname=foo host=baz')
         self.assertEqual(self.args[1], f)
         self.assertEqual(self.args[2], False)
 
-        psycopg2.connect("dbname=foo bar=baz", connection_factory=f)
-        self.assertEqual(self.args[0], 'dbname=foo bar=baz')
+        psycopg2.connect("dbname=foo host=baz", connection_factory=f)
+        self.assertDsnEqual(self.args[0], 'dbname=foo host=baz')
         self.assertEqual(self.args[1], f)
         self.assertEqual(self.args[2], False)
 
     def test_async(self):
-        psycopg2.connect(database='foo', bar='baz', async=1)
-        self.assertEqual(self.args[0], 'dbname=foo bar=baz')
+        psycopg2.connect(database='foo', host='baz', async_=1)
+        self.assertDsnEqual(self.args[0], 'dbname=foo host=baz')
         self.assertEqual(self.args[1], None)
         self.assert_(self.args[2])
 
-        psycopg2.connect("dbname=foo bar=baz", async=True)
-        self.assertEqual(self.args[0], 'dbname=foo bar=baz')
+        psycopg2.connect("dbname=foo host=baz", async_=True)
+        self.assertDsnEqual(self.args[0], 'dbname=foo host=baz')
         self.assertEqual(self.args[1], None)
         self.assert_(self.args[2])
 
+    def test_int_port_param(self):
+        psycopg2.connect(database='sony', port=6543)
+        dsn = " %s " % self.args[0]
+        self.assert_(" dbname=sony " in dsn, dsn)
+        self.assert_(" port=6543 " in dsn, dsn)
+
     def test_empty_param(self):
         psycopg2.connect(database='sony', password='')
-        self.assertEqual(self.args[0], "dbname=sony password=''")
+        self.assertDsnEqual(self.args[0], "dbname=sony password=''")
 
     def test_escape(self):
         psycopg2.connect(database='hello world')
@@ -127,13 +139,12 @@ class ConnectTestCase(unittest.TestCase):
         psycopg2.connect(database=r"\every thing'")
         self.assertEqual(self.args[0], r"dbname='\\every thing\''")
 
-    def test_no_kwargs_swallow(self):
-        self.assertRaises(TypeError,
-            psycopg2.connect, 'dbname=foo', database='foo')
-        self.assertRaises(TypeError,
-            psycopg2.connect, 'dbname=foo', user='postgres')
-        self.assertRaises(TypeError,
-            psycopg2.connect, 'dbname=foo', no_such_param='meh')
+    def test_params_merging(self):
+        psycopg2.connect('dbname=foo', database='bar')
+        self.assertEqual(self.args[0], 'dbname=bar')
+
+        psycopg2.connect('dbname=foo', user='postgres')
+        self.assertDsnEqual(self.args[0], 'dbname=foo user=postgres')
 
 
 class ExceptionsTestCase(ConnectingTestCase):
@@ -141,7 +152,7 @@ class ExceptionsTestCase(ConnectingTestCase):
         cur = self.conn.cursor()
         try:
             cur.execute("select * from nonexist")
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             e = exc
 
         self.assertEqual(e.pgcode, '42P01')
@@ -152,7 +163,7 @@ class ExceptionsTestCase(ConnectingTestCase):
         cur = self.conn.cursor()
         try:
             cur.execute("select * from nonexist")
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             e = exc
 
         diag = e.diag
@@ -171,7 +182,7 @@ class ExceptionsTestCase(ConnectingTestCase):
         cur = self.conn.cursor()
         try:
             cur.execute("select * from nonexist")
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             e = exc
 
         self.assertEqual(e.diag.sqlstate, '42P01')
@@ -185,7 +196,7 @@ class ExceptionsTestCase(ConnectingTestCase):
             cur = self.conn.cursor()
             try:
                 cur.execute("select * from nonexist")
-            except psycopg2.Error, exc:
+            except psycopg2.Error as exc:
                 return cur, exc
 
         cur, e = tmp()
@@ -199,7 +210,8 @@ class ExceptionsTestCase(ConnectingTestCase):
         self.assertEqual(diag.sqlstate, '42P01')
 
         del diag
-        gc.collect(); gc.collect()
+        gc.collect()
+        gc.collect()
         assert(w() is None)
 
     @skip_copy_if_green
@@ -209,7 +221,7 @@ class ExceptionsTestCase(ConnectingTestCase):
         cur = self.conn.cursor()
         try:
             cur.copy_to(f, 'nonexist')
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             diag = exc.diag
 
         self.assertEqual(diag.sqlstate, '42P01')
@@ -218,14 +230,14 @@ class ExceptionsTestCase(ConnectingTestCase):
         cur = self.conn.cursor()
         try:
             cur.execute("l'acqua e' poca e 'a papera nun galleggia")
-        except Exception, exc:
+        except Exception as exc:
             diag1 = exc.diag
 
         self.conn.rollback()
 
         try:
             cur.execute("select level from water where ducks > 1")
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             diag2 = exc.diag
 
         self.assertEqual(diag1.sqlstate, '42601')
@@ -242,7 +254,7 @@ class ExceptionsTestCase(ConnectingTestCase):
         cur.execute("insert into test_deferred values (1,2)")
         try:
             self.conn.commit()
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             e = exc
         self.assertEqual(e.diag.sqlstate, '23503')
 
@@ -255,7 +267,7 @@ class ExceptionsTestCase(ConnectingTestCase):
             )""")
         try:
             cur.execute("insert into test_exc values(2)")
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             e = exc
         self.assertEqual(e.pgcode, '23514')
         self.assertEqual(e.diag.schema_name[:7], "pg_temp")
@@ -270,7 +282,7 @@ class ExceptionsTestCase(ConnectingTestCase):
         cur = self.conn.cursor()
         try:
             cur.execute("select * from nonexist")
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             e = exc
 
         e1 = pickle.loads(pickle.dumps(e))
@@ -285,7 +297,7 @@ class ExceptionsTestCase(ConnectingTestCase):
         import pickle
         try:
             psycopg2.connect('dbname=nosuchdatabasemate')
-        except psycopg2.Error, exc:
+        except psycopg2.Error as exc:
             e = exc
 
         e1 = pickle.loads(pickle.dumps(e))
@@ -295,6 +307,37 @@ class ExceptionsTestCase(ConnectingTestCase):
         self.assert_(e1.cursor is None)
 
 
+class TestExtensionModule(unittest.TestCase):
+    @slow
+    def test_import_internal(self):
+        # check that the internal package can be imported "naked"
+        # we may break this property if there is a compelling reason to do so,
+        # however having it allows for some import juggling such as the one
+        # required in ticket #201.
+        pkgdir = os.path.dirname(psycopg2.__file__)
+        pardir = os.path.dirname(pkgdir)
+        self.assert_(pardir in sys.path)
+        script = ("""
+import sys
+sys.path.remove(%r)
+sys.path.insert(0, %r)
+import _psycopg
+""" % (pardir, pkgdir))
+
+        proc = Popen([sys.executable, '-c', script_to_py3(script)])
+        proc.communicate()
+        self.assertEqual(0, proc.returncode)
+
+
+class TestVersionDiscovery(unittest.TestCase):
+    def test_libpq_version(self):
+        self.assertTrue(type(psycopg2.__libpq_version__) is int)
+        try:
+            self.assertTrue(type(psycopg2.extensions.libpq_version()) is int)
+        except psycopg2.NotSupportedError:
+            self.assertTrue(psycopg2.__libpq_version__ < 90100)
+
+
 def test_suite():
     return unittest.TestLoader().loadTestsFromName(__name__)
 
