@@ -13,27 +13,24 @@
 # ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 # FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 # License for more details.
+from __future__ import with_statement
 
 import re
 import sys
+import warnings
 from decimal import Decimal
 from datetime import date, datetime
 from functools import wraps
+from pickle import dumps, loads
 
-from testutils import unittest, skip_if_no_uuid, skip_before_postgres
-from testutils import ConnectingTestCase, decorate_all_tests
+from testutils import (unittest, skip_if_no_uuid, skip_before_postgres,
+    ConnectingTestCase, decorate_all_tests, py3_raises_typeerror, slow)
 
 import psycopg2
 import psycopg2.extras
-from psycopg2.extensions import b
+import psycopg2.extensions as ext
 
 
-def filter_scs(conn, s):
-    if conn.get_parameter_status("standard_conforming_strings") == 'off':
-        return s
-    else:
-        return s.replace(b("E'"), b("'"))
-
 class TypesExtrasTests(ConnectingTestCase):
     """Test that all type conversions are working."""
 
@@ -57,7 +54,8 @@ class TypesExtrasTests(ConnectingTestCase):
     def testUUIDARRAY(self):
         import uuid
         psycopg2.extras.register_uuid()
-        u = [uuid.UUID('9c6d5a77-7256-457e-9461-347b4358e350'), uuid.UUID('9c6d5a77-7256-457e-9461-347b4358e352')]
+        u = [uuid.UUID('9c6d5a77-7256-457e-9461-347b4358e350'),
+             uuid.UUID('9c6d5a77-7256-457e-9461-347b4358e352')]
         s = self.execute("SELECT %s AS foo", (u,))
         self.failUnless(u == s)
         # array with a NULL element
@@ -72,7 +70,10 @@ class TypesExtrasTests(ConnectingTestCase):
         self.failUnless(type(s) == list and len(s) == 0)
 
     def testINET(self):
-        psycopg2.extras.register_inet()
+        with warnings.catch_warnings():
+            warnings.simplefilter('ignore', DeprecationWarning)
+            psycopg2.extras.register_inet()
+
         i = psycopg2.extras.Inet("192.168.1.0/24")
         s = self.execute("SELECT %s AS foo", (i,))
         self.failUnless(i.addr == s.addr)
@@ -81,7 +82,10 @@ class TypesExtrasTests(ConnectingTestCase):
         self.failUnless(s is None)
 
     def testINETARRAY(self):
-        psycopg2.extras.register_inet()
+        with warnings.catch_warnings():
+            warnings.simplefilter('ignore', DeprecationWarning)
+            psycopg2.extras.register_inet()
+
         i = psycopg2.extras.Inet("192.168.1.0/24")
         s = self.execute("SELECT %s AS foo", ([i],))
         self.failUnless(i.addr == s[0].addr)
@@ -94,27 +98,29 @@ class TypesExtrasTests(ConnectingTestCase):
         i = Inet("192.168.1.0/24")
         a = psycopg2.extensions.adapt(i)
         a.prepare(self.conn)
-        self.assertEqual(
-            filter_scs(self.conn, b("E'192.168.1.0/24'::inet")),
-            a.getquoted())
+        self.assertQuotedEqual(a.getquoted(), b"'192.168.1.0/24'::inet")
 
         # adapts ok with unicode too
         i = Inet(u"192.168.1.0/24")
         a = psycopg2.extensions.adapt(i)
         a.prepare(self.conn)
-        self.assertEqual(
-            filter_scs(self.conn, b("E'192.168.1.0/24'::inet")),
-            a.getquoted())
+        self.assertQuotedEqual(a.getquoted(), b"'192.168.1.0/24'::inet")
 
     def test_adapt_fail(self):
-        class Foo(object): pass
+        class Foo(object):
+            pass
         self.assertRaises(psycopg2.ProgrammingError,
-            psycopg2.extensions.adapt, Foo(), psycopg2.extensions.ISQLQuote, None)
+            psycopg2.extensions.adapt, Foo(), ext.ISQLQuote, None)
         try:
-            psycopg2.extensions.adapt(Foo(), psycopg2.extensions.ISQLQuote, None)
-        except psycopg2.ProgrammingError, err:
+            psycopg2.extensions.adapt(Foo(), ext.ISQLQuote, None)
+        except psycopg2.ProgrammingError as err:
             self.failUnless(str(err) == "can't adapt type 'Foo'")
 
+    def test_point_array(self):
+        # make sure a point array is never casted to a float array,
+        # see https://github.com/psycopg/psycopg2/issues/613
+        s = self.execute("""SELECT '{"(1,2)","(3,4)"}' AS foo""")
+        self.failUnless(s == """{"(1,2)","(3,4)"}""")
 
 def skip_if_no_hstore(f):
     @wraps(f)
@@ -127,6 +133,7 @@ def skip_if_no_hstore(f):
 
     return skip_if_no_hstore_
 
+
 class HstoreTestCase(ConnectingTestCase):
     def test_adapt_8(self):
         if self.conn.server_version >= 90000:
@@ -142,17 +149,17 @@ class HstoreTestCase(ConnectingTestCase):
         a.prepare(self.conn)
         q = a.getquoted()
 
-        self.assert_(q.startswith(b("((")), q)
-        ii = q[1:-1].split(b("||"))
+        self.assert_(q.startswith(b"(("), q)
+        ii = q[1:-1].split(b"||")
         ii.sort()
 
         self.assertEqual(len(ii), len(o))
-        self.assertEqual(ii[0], filter_scs(self.conn, b("(E'a' => E'1')")))
-        self.assertEqual(ii[1], filter_scs(self.conn, b("(E'b' => E'''')")))
-        self.assertEqual(ii[2], filter_scs(self.conn, b("(E'c' => NULL)")))
+        self.assertQuotedEqual(ii[0], b"('a' => '1')")
+        self.assertQuotedEqual(ii[1], b"('b' => '''')")
+        self.assertQuotedEqual(ii[2], b"('c' => NULL)")
         if 'd' in o:
             encc = u'\xe0'.encode(psycopg2.extensions.encodings[self.conn.encoding])
-            self.assertEqual(ii[3], filter_scs(self.conn, b("(E'd' => E'") + encc + b("')")))
+            self.assertQuotedEqual(ii[3], b"('d' => '" + encc + b"')")
 
     def test_adapt_9(self):
         if self.conn.server_version < 90000:
@@ -168,24 +175,25 @@ class HstoreTestCase(ConnectingTestCase):
         a.prepare(self.conn)
         q = a.getquoted()
 
-        m = re.match(b(r'hstore\(ARRAY\[([^\]]+)\], ARRAY\[([^\]]+)\]\)'), q)
+        m = re.match(br'hstore\(ARRAY\[([^\]]+)\], ARRAY\[([^\]]+)\]\)', q)
         self.assert_(m, repr(q))
 
-        kk = m.group(1).split(b(", "))
-        vv = m.group(2).split(b(", "))
+        kk = m.group(1).split(b", ")
+        vv = m.group(2).split(b", ")
         ii = zip(kk, vv)
         ii.sort()
 
-        def f(*args):
-            return tuple([filter_scs(self.conn, s) for s in args])
-
         self.assertEqual(len(ii), len(o))
-        self.assertEqual(ii[0], f(b("E'a'"), b("E'1'")))
-        self.assertEqual(ii[1], f(b("E'b'"), b("E''''")))
-        self.assertEqual(ii[2], f(b("E'c'"), b("NULL")))
+        self.assertQuotedEqual(ii[0][0], b"'a'")
+        self.assertQuotedEqual(ii[0][1], b"'1'")
+        self.assertQuotedEqual(ii[1][0], b"'b'")
+        self.assertQuotedEqual(ii[1][1], b"''''")
+        self.assertQuotedEqual(ii[2][0], b"'c'")
+        self.assertQuotedEqual(ii[2][1], b"NULL")
         if 'd' in o:
             encc = u'\xe0'.encode(psycopg2.extensions.encodings[self.conn.encoding])
-            self.assertEqual(ii[3], f(b("E'd'"), b("E'") + encc + b("'")))
+            self.assertQuotedEqual(ii[3][0], b"'d'")
+            self.assertQuotedEqual(ii[3][1], b"'" + encc + b"'")
 
     def test_parse(self):
         from psycopg2.extras import HstoreAdapter
@@ -196,7 +204,7 @@ class HstoreTestCase(ConnectingTestCase):
         ok(None, None)
         ok('', {})
         ok('"a"=>"1", "b"=>"2"', {'a': '1', 'b': '2'})
-        ok('"a"  => "1" ,"b"  =>  "2"', {'a': '1', 'b': '2'})
+        ok('"a"  => "1" , "b"  =>  "2"', {'a': '1', 'b': '2'})
         ok('"a"=>NULL, "b"=>"2"', {'a': None, 'b': '2'})
         ok(r'"a"=>"\"", "\""=>"2"', {'a': '"', '"': '2'})
         ok('"a"=>"\'", "\'"=>"2"', {'a': "'", "'": '2'})
@@ -399,7 +407,9 @@ class HstoreTestCase(ConnectingTestCase):
         from psycopg2.extras import register_hstore
         register_hstore(None, globally=True, oid=oid, array_oid=aoid)
         try:
-            cur.execute("select null::hstore, ''::hstore, 'a => b'::hstore, '{a=>b}'::hstore[]")
+            cur.execute("""
+                select null::hstore, ''::hstore,
+                'a => b'::hstore, '{a=>b}'::hstore[]""")
             t = cur.fetchone()
             self.assert_(t[0] is None)
             self.assertEqual(t[1], {})
@@ -446,23 +456,26 @@ def skip_if_no_composite(f):
 
     return skip_if_no_composite_
 
+
 class AdaptTypeTestCase(ConnectingTestCase):
     @skip_if_no_composite
     def test_none_in_record(self):
         curs = self.conn.cursor()
         s = curs.mogrify("SELECT %s;", [(42, None)])
-        self.assertEqual(b("SELECT (42, NULL);"), s)
+        self.assertEqual(b"SELECT (42, NULL);", s)
         curs.execute("SELECT %s;", [(42, None)])
         d = curs.fetchone()[0]
         self.assertEqual("(42,)", d)
 
     def test_none_fast_path(self):
         # the None adapter is not actually invoked in regular adaptation
-        ext = psycopg2.extensions
 
         class WonkyAdapter(object):
-            def __init__(self, obj): pass
-            def getquoted(self): return "NOPE!"
+            def __init__(self, obj):
+                pass
+
+            def getquoted(self):
+                return "NOPE!"
 
         curs = self.conn.cursor()
 
@@ -472,13 +485,14 @@ class AdaptTypeTestCase(ConnectingTestCase):
             self.assertEqual(ext.adapt(None).getquoted(), "NOPE!")
 
             s = curs.mogrify("SELECT %s;", (None,))
-            self.assertEqual(b("SELECT NULL;"), s)
+            self.assertEqual(b"SELECT NULL;", s)
 
         finally:
             ext.register_adapter(type(None), orig_adapter)
 
     def test_tokenization(self):
         from psycopg2.extras import CompositeCaster
+
         def ok(s, v):
             self.assertEqual(CompositeCaster.tokenize(s), v)
 
@@ -517,26 +531,26 @@ class AdaptTypeTestCase(ConnectingTestCase):
         self.assertEqual(t.oid, oid)
         self.assert_(issubclass(t.type, tuple))
         self.assertEqual(t.attnames, ['anint', 'astring', 'adate'])
-        self.assertEqual(t.atttypes, [23,25,1082])
+        self.assertEqual(t.atttypes, [23, 25, 1082])
 
         curs = self.conn.cursor()
-        r = (10, 'hello', date(2011,1,2))
+        r = (10, 'hello', date(2011, 1, 2))
         curs.execute("select %s::type_isd;", (r,))
         v = curs.fetchone()[0]
         self.assert_(isinstance(v, t.type))
         self.assertEqual(v[0], 10)
         self.assertEqual(v[1], "hello")
-        self.assertEqual(v[2], date(2011,1,2))
+        self.assertEqual(v[2], date(2011, 1, 2))
 
         try:
-            from collections import namedtuple
+            from collections import namedtuple          # noqa
         except ImportError:
             pass
         else:
             self.assert_(t.type is not tuple)
             self.assertEqual(v.anint, 10)
             self.assertEqual(v.astring, "hello")
-            self.assertEqual(v.adate, date(2011,1,2))
+            self.assertEqual(v.adate, date(2011, 1, 2))
 
     @skip_if_no_composite
     def test_empty_string(self):
@@ -572,14 +586,14 @@ class AdaptTypeTestCase(ConnectingTestCase):
         psycopg2.extras.register_composite("type_r_ft", self.conn)
 
         curs = self.conn.cursor()
-        r = (0.25, (date(2011,1,2), (42, "hello")))
+        r = (0.25, (date(2011, 1, 2), (42, "hello")))
         curs.execute("select %s::type_r_ft;", (r,))
         v = curs.fetchone()[0]
 
         self.assertEqual(r, v)
 
         try:
-            from collections import namedtuple
+            from collections import namedtuple              # noqa
         except ImportError:
             pass
         else:
@@ -593,7 +607,7 @@ class AdaptTypeTestCase(ConnectingTestCase):
         curs2 = self.conn.cursor()
         psycopg2.extras.register_composite("type_ii", curs1)
         curs1.execute("select (1,2)::type_ii")
-        self.assertEqual(curs1.fetchone()[0], (1,2))
+        self.assertEqual(curs1.fetchone()[0], (1, 2))
         curs2.execute("select (1,2)::type_ii")
         self.assertEqual(curs2.fetchone()[0], "(1,2)")
 
@@ -608,7 +622,7 @@ class AdaptTypeTestCase(ConnectingTestCase):
             curs1 = conn1.cursor()
             curs2 = conn2.cursor()
             curs1.execute("select (1,2)::type_ii")
-            self.assertEqual(curs1.fetchone()[0], (1,2))
+            self.assertEqual(curs1.fetchone()[0], (1, 2))
             curs2.execute("select (1,2)::type_ii")
             self.assertEqual(curs2.fetchone()[0], "(1,2)")
         finally:
@@ -627,9 +641,9 @@ class AdaptTypeTestCase(ConnectingTestCase):
                 curs1 = conn1.cursor()
                 curs2 = conn2.cursor()
                 curs1.execute("select (1,2)::type_ii")
-                self.assertEqual(curs1.fetchone()[0], (1,2))
+                self.assertEqual(curs1.fetchone()[0], (1, 2))
                 curs2.execute("select (1,2)::type_ii")
-                self.assertEqual(curs2.fetchone()[0], (1,2))
+                self.assertEqual(curs2.fetchone()[0], (1, 2))
             finally:
                 # drop the registered typecasters to help the refcounting
                 # script to return precise values.
@@ -659,30 +673,30 @@ class AdaptTypeTestCase(ConnectingTestCase):
             "typens.typens_ii", self.conn)
         self.assertEqual(t.schema, 'typens')
         curs.execute("select (4,8)::typens.typens_ii")
-        self.assertEqual(curs.fetchone()[0], (4,8))
+        self.assertEqual(curs.fetchone()[0], (4, 8))
 
     @skip_if_no_composite
     @skip_before_postgres(8, 4)
     def test_composite_array(self):
-        oid = self._create_type("type_isd",
+        self._create_type("type_isd",
             [('anint', 'integer'), ('astring', 'text'), ('adate', 'date')])
 
         t = psycopg2.extras.register_composite("type_isd", self.conn)
 
         curs = self.conn.cursor()
-        r1 = (10, 'hello', date(2011,1,2))
-        r2 = (20, 'world', date(2011,1,3))
+        r1 = (10, 'hello', date(2011, 1, 2))
+        r2 = (20, 'world', date(2011, 1, 3))
         curs.execute("select %s::type_isd[];", ([r1, r2],))
         v = curs.fetchone()[0]
         self.assertEqual(len(v), 2)
         self.assert_(isinstance(v[0], t.type))
         self.assertEqual(v[0][0], 10)
         self.assertEqual(v[0][1], "hello")
-        self.assertEqual(v[0][2], date(2011,1,2))
+        self.assertEqual(v[0][2], date(2011, 1, 2))
         self.assert_(isinstance(v[1], t.type))
         self.assertEqual(v[1][0], 20)
         self.assertEqual(v[1][1], "world")
-        self.assertEqual(v[1][2], date(2011,1,3))
+        self.assertEqual(v[1][2], date(2011, 1, 3))
 
     @skip_if_no_composite
     def test_wrong_schema(self):
@@ -694,6 +708,7 @@ class AdaptTypeTestCase(ConnectingTestCase):
         curs.execute("select (1,2)::type_ii")
         self.assertRaises(psycopg2.DataError, curs.fetchone)
 
+    @slow
     @skip_if_no_composite
     @skip_before_postgres(8, 4)
     def test_from_tables(self):
@@ -750,7 +765,7 @@ class AdaptTypeTestCase(ConnectingTestCase):
             register_composite('type_ii', conn)
             curs = conn.cursor()
             curs.execute("select '(1,2)'::type_ii as x")
-            self.assertEqual(curs.fetchone()['x'], (1,2))
+            self.assertEqual(curs.fetchone()['x'], (1, 2))
         finally:
             conn.close()
 
@@ -759,7 +774,7 @@ class AdaptTypeTestCase(ConnectingTestCase):
             curs = conn.cursor()
             register_composite('type_ii', conn)
             curs.execute("select '(1,2)'::type_ii as x")
-            self.assertEqual(curs.fetchone()['x'], (1,2))
+            self.assertEqual(curs.fetchone()['x'], (1, 2))
         finally:
             conn.close()
 
@@ -780,13 +795,13 @@ class AdaptTypeTestCase(ConnectingTestCase):
         self.assertEqual(t.oid, oid)
 
         curs = self.conn.cursor()
-        r = (10, 'hello', date(2011,1,2))
+        r = (10, 'hello', date(2011, 1, 2))
         curs.execute("select %s::type_isd;", (r,))
         v = curs.fetchone()[0]
         self.assert_(isinstance(v, dict))
         self.assertEqual(v['anint'], 10)
         self.assertEqual(v['astring'], "hello")
-        self.assertEqual(v['adate'], date(2011,1,2))
+        self.assertEqual(v['adate'], date(2011, 1, 2))
 
     def _create_type(self, name, fields):
         curs = self.conn.cursor()
@@ -823,6 +838,7 @@ def skip_if_json_module(f):
 
     return skip_if_json_module_
 
+
 def skip_if_no_json_module(f):
     """Skip a test if no Python json module is available"""
     @wraps(f)
@@ -834,6 +850,7 @@ def skip_if_no_json_module(f):
 
     return skip_if_no_json_module_
 
+
 def skip_if_no_json_type(f):
     """Skip a test if PostgreSQL json type is not available"""
     @wraps(f)
@@ -847,6 +864,7 @@ def skip_if_no_json_type(f):
 
     return skip_if_no_json_type_
 
+
 class JsonTestCase(ConnectingTestCase):
     @skip_if_json_module
     def test_module_not_available(self):
@@ -856,6 +874,7 @@ class JsonTestCase(ConnectingTestCase):
     @skip_if_json_module
     def test_customizable_with_module_not_available(self):
         from psycopg2.extras import Json
+
         class MyJson(Json):
             def dumps(self, obj):
                 assert obj is None
@@ -868,11 +887,11 @@ class JsonTestCase(ConnectingTestCase):
         from psycopg2.extras import json, Json
 
         objs = [None, "te'xt", 123, 123.45,
-            u'\xe0\u20ac', ['a', 100], {'a': 100} ]
+            u'\xe0\u20ac', ['a', 100], {'a': 100}]
 
         curs = self.conn.cursor()
         for obj in enumerate(objs):
-            self.assertEqual(curs.mogrify("%s", (Json(obj),)),
+            self.assertQuotedEqual(curs.mogrify("%s", (Json(obj),)),
                 psycopg2.extensions.QuotedString(json.dumps(obj)).getquoted())
 
     @skip_if_no_json_module
@@ -887,9 +906,11 @@ class JsonTestCase(ConnectingTestCase):
 
         curs = self.conn.cursor()
         obj = Decimal('123.45')
-        dumps = lambda obj: json.dumps(obj, cls=DecimalEncoder)
-        self.assertEqual(curs.mogrify("%s", (Json(obj, dumps=dumps),)),
-            b("'123.45'"))
+
+        def dumps(obj):
+            return json.dumps(obj, cls=DecimalEncoder)
+        self.assertQuotedEqual(curs.mogrify("%s", (Json(obj, dumps=dumps),)),
+            b"'123.45'")
 
     @skip_if_no_json_module
     def test_adapt_subclass(self):
@@ -907,8 +928,7 @@ class JsonTestCase(ConnectingTestCase):
 
         curs = self.conn.cursor()
         obj = Decimal('123.45')
-        self.assertEqual(curs.mogrify("%s", (MyJson(obj),)),
-            b("'123.45'"))
+        self.assertQuotedEqual(curs.mogrify("%s", (MyJson(obj),)), b"'123.45'")
 
     @skip_if_no_json_module
     def test_register_on_dict(self):
@@ -918,11 +938,10 @@ class JsonTestCase(ConnectingTestCase):
         try:
             curs = self.conn.cursor()
             obj = {'a': 123}
-            self.assertEqual(curs.mogrify("%s", (obj,)),
-                b("""'{"a": 123}'"""))
+            self.assertQuotedEqual(
+                curs.mogrify("%s", (obj,)), b"""'{"a": 123}'""")
         finally:
-           del psycopg2.extensions.adapters[dict, psycopg2.extensions.ISQLQuote]
-
+            del psycopg2.extensions.adapters[dict, ext.ISQLQuote]
 
     def test_type_not_available(self):
         curs = self.conn.cursor()
@@ -982,7 +1001,9 @@ class JsonTestCase(ConnectingTestCase):
     @skip_if_no_json_type
     def test_loads(self):
         json = psycopg2.extras.json
-        loads = lambda x: json.loads(x, parse_float=Decimal)
+
+        def loads(s):
+            return json.loads(s, parse_float=Decimal)
         psycopg2.extras.register_json(self.conn, loads=loads)
         curs = self.conn.cursor()
         curs.execute("""select '{"a": 100.0, "b": null}'::json""")
@@ -998,7 +1019,9 @@ class JsonTestCase(ConnectingTestCase):
 
         old = psycopg2.extensions.string_types.get(114)
         olda = psycopg2.extensions.string_types.get(199)
-        loads = lambda x: psycopg2.extras.json.loads(x, parse_float=Decimal)
+
+        def loads(s):
+            return psycopg2.extras.json.loads(s, parse_float=Decimal)
         try:
             new, newa = psycopg2.extras.register_json(
                 loads=loads, oid=oid, array_oid=array_oid)
@@ -1020,7 +1043,8 @@ class JsonTestCase(ConnectingTestCase):
     def test_register_default(self):
         curs = self.conn.cursor()
 
-        loads = lambda x: psycopg2.extras.json.loads(x, parse_float=Decimal)
+        def loads(s):
+            return psycopg2.extras.json.loads(s, parse_float=Decimal)
         psycopg2.extras.register_default_json(curs, loads=loads)
 
         curs.execute("""select '{"a": 100.0, "b": null}'::json""")
@@ -1055,6 +1079,133 @@ class JsonTestCase(ConnectingTestCase):
         self.assertEqual(data['a'], 100)
         self.assertEqual(data['b'], None)
 
+    @skip_if_no_json_module
+    def test_str(self):
+        snowman = u"\u2603"
+        obj = {'a': [1, 2, snowman]}
+        j = psycopg2.extensions.adapt(psycopg2.extras.Json(obj))
+        s = str(j)
+        self.assert_(isinstance(s, str))
+        # no pesky b's
+        self.assert_(s.startswith("'"))
+        self.assert_(s.endswith("'"))
+
+    @skip_if_no_json_module
+    @skip_before_postgres(8, 2)
+    def test_scs(self):
+        cnn_on = self.connect(options="-c standard_conforming_strings=on")
+        cur_on = cnn_on.cursor()
+        self.assertEqual(
+            cur_on.mogrify("%s", [psycopg2.extras.Json({'a': '"'})]),
+            b'\'{"a": "\\""}\'')
+
+        cnn_off = self.connect(options="-c standard_conforming_strings=off")
+        cur_off = cnn_off.cursor()
+        self.assertEqual(
+            cur_off.mogrify("%s", [psycopg2.extras.Json({'a': '"'})]),
+            b'E\'{"a": "\\\\""}\'')
+
+        self.assertEqual(
+            cur_on.mogrify("%s", [psycopg2.extras.Json({'a': '"'})]),
+            b'\'{"a": "\\""}\'')
+
+
+def skip_if_no_jsonb_type(f):
+    return skip_before_postgres(9, 4)(f)
+
+
+class JsonbTestCase(ConnectingTestCase):
+    @staticmethod
+    def myloads(s):
+        import json
+        rv = json.loads(s)
+        rv['test'] = 1
+        return rv
+
+    def test_default_cast(self):
+        curs = self.conn.cursor()
+
+        curs.execute("""select '{"a": 100.0, "b": null}'::jsonb""")
+        self.assertEqual(curs.fetchone()[0], {'a': 100.0, 'b': None})
+
+        curs.execute("""select array['{"a": 100.0, "b": null}']::jsonb[]""")
+        self.assertEqual(curs.fetchone()[0], [{'a': 100.0, 'b': None}])
+
+    def test_register_on_connection(self):
+        psycopg2.extras.register_json(self.conn, loads=self.myloads, name='jsonb')
+        curs = self.conn.cursor()
+        curs.execute("""select '{"a": 100.0, "b": null}'::jsonb""")
+        self.assertEqual(curs.fetchone()[0], {'a': 100.0, 'b': None, 'test': 1})
+
+    def test_register_on_cursor(self):
+        curs = self.conn.cursor()
+        psycopg2.extras.register_json(curs, loads=self.myloads, name='jsonb')
+        curs.execute("""select '{"a": 100.0, "b": null}'::jsonb""")
+        self.assertEqual(curs.fetchone()[0], {'a': 100.0, 'b': None, 'test': 1})
+
+    def test_register_globally(self):
+        old = psycopg2.extensions.string_types.get(3802)
+        olda = psycopg2.extensions.string_types.get(3807)
+        try:
+            new, newa = psycopg2.extras.register_json(self.conn,
+                loads=self.myloads, globally=True, name='jsonb')
+            curs = self.conn.cursor()
+            curs.execute("""select '{"a": 100.0, "b": null}'::jsonb""")
+            self.assertEqual(curs.fetchone()[0], {'a': 100.0, 'b': None, 'test': 1})
+        finally:
+            psycopg2.extensions.string_types.pop(new.values[0])
+            psycopg2.extensions.string_types.pop(newa.values[0])
+            if old:
+                psycopg2.extensions.register_type(old)
+            if olda:
+                psycopg2.extensions.register_type(olda)
+
+    def test_loads(self):
+        json = psycopg2.extras.json
+
+        def loads(s):
+            return json.loads(s, parse_float=Decimal)
+
+        psycopg2.extras.register_json(self.conn, loads=loads, name='jsonb')
+        curs = self.conn.cursor()
+        curs.execute("""select '{"a": 100.0, "b": null}'::jsonb""")
+        data = curs.fetchone()[0]
+        self.assert_(isinstance(data['a'], Decimal))
+        self.assertEqual(data['a'], Decimal('100.0'))
+        # sure we are not manling json too?
+        curs.execute("""select '{"a": 100.0, "b": null}'::json""")
+        data = curs.fetchone()[0]
+        self.assert_(isinstance(data['a'], float))
+        self.assertEqual(data['a'], 100.0)
+
+    def test_register_default(self):
+        curs = self.conn.cursor()
+
+        def loads(s):
+            return psycopg2.extras.json.loads(s, parse_float=Decimal)
+
+        psycopg2.extras.register_default_jsonb(curs, loads=loads)
+
+        curs.execute("""select '{"a": 100.0, "b": null}'::jsonb""")
+        data = curs.fetchone()[0]
+        self.assert_(isinstance(data['a'], Decimal))
+        self.assertEqual(data['a'], Decimal('100.0'))
+
+        curs.execute("""select array['{"a": 100.0, "b": null}']::jsonb[]""")
+        data = curs.fetchone()[0]
+        self.assert_(isinstance(data[0]['a'], Decimal))
+        self.assertEqual(data[0]['a'], Decimal('100.0'))
+
+    def test_null(self):
+        curs = self.conn.cursor()
+        curs.execute("""select NULL::jsonb""")
+        self.assertEqual(curs.fetchone()[0], None)
+        curs.execute("""select NULL::jsonb[]""")
+        self.assertEqual(curs.fetchone()[0], None)
+
+decorate_all_tests(JsonbTestCase, skip_if_no_json_module)
+decorate_all_tests(JsonbTestCase, skip_if_no_jsonb_type)
+
 
 class RangeTestCase(unittest.TestCase):
     def test_noparam(self):
@@ -1098,7 +1249,7 @@ class RangeTestCase(unittest.TestCase):
                 ('[)', True, False),
                 ('(]', False, True),
                 ('()', False, False),
-                ('[]', True, True),]:
+                ('[]', True, True)]:
             r = Range(10, 20, bounds)
             self.assertEqual(r.lower, 10)
             self.assertEqual(r.upper, 20)
@@ -1192,11 +1343,11 @@ class RangeTestCase(unittest.TestCase):
         self.assert_(not Range(empty=True))
 
     def test_eq_hash(self):
-        from psycopg2.extras import Range
         def assert_equal(r1, r2):
             self.assert_(r1 == r2)
             self.assert_(hash(r1) == hash(r2))
 
+        from psycopg2.extras import Range
         assert_equal(Range(empty=True), Range(empty=True))
         assert_equal(Range(), Range())
         assert_equal(Range(10, None), Range(10, None))
@@ -1212,12 +1363,95 @@ class RangeTestCase(unittest.TestCase):
         assert_not_equal(Range(10, 20), Range(11, 20))
         assert_not_equal(Range(10, 20, '[)'), Range(10, 20, '[]'))
 
-    def test_not_ordered(self):
+    def test_eq_wrong_type(self):
         from psycopg2.extras import Range
-        self.assertRaises(TypeError, lambda: Range(empty=True) < Range(0,4))
-        self.assertRaises(TypeError, lambda: Range(1,2) > Range(0,4))
-        self.assertRaises(TypeError, lambda: Range(1,2) <= Range())
-        self.assertRaises(TypeError, lambda: Range(1,2) >= Range())
+        self.assertNotEqual(Range(10, 20), ())
+
+    def test_eq_subclass(self):
+        from psycopg2.extras import Range, NumericRange
+
+        class IntRange(NumericRange):
+            pass
+
+        class PositiveIntRange(IntRange):
+            pass
+
+        self.assertEqual(Range(10, 20), IntRange(10, 20))
+        self.assertEqual(PositiveIntRange(10, 20), IntRange(10, 20))
+
+    # as the postgres docs describe for the server-side stuff,
+    # ordering is rather arbitrary, but will remain stable
+    # and consistent.
+
+    def test_lt_ordering(self):
+        from psycopg2.extras import Range
+        self.assert_(Range(empty=True) < Range(0, 4))
+        self.assert_(not Range(1, 2) < Range(0, 4))
+        self.assert_(Range(0, 4) < Range(1, 2))
+        self.assert_(not Range(1, 2) < Range())
+        self.assert_(Range() < Range(1, 2))
+        self.assert_(not Range(1) < Range(upper=1))
+        self.assert_(not Range() < Range())
+        self.assert_(not Range(empty=True) < Range(empty=True))
+        self.assert_(not Range(1, 2) < Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(1 < Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(not Range(1, 2) < 1)
+
+    def test_gt_ordering(self):
+        from psycopg2.extras import Range
+        self.assert_(not Range(empty=True) > Range(0, 4))
+        self.assert_(Range(1, 2) > Range(0, 4))
+        self.assert_(not Range(0, 4) > Range(1, 2))
+        self.assert_(Range(1, 2) > Range())
+        self.assert_(not Range() > Range(1, 2))
+        self.assert_(Range(1) > Range(upper=1))
+        self.assert_(not Range() > Range())
+        self.assert_(not Range(empty=True) > Range(empty=True))
+        self.assert_(not Range(1, 2) > Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(not 1 > Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(Range(1, 2) > 1)
+
+    def test_le_ordering(self):
+        from psycopg2.extras import Range
+        self.assert_(Range(empty=True) <= Range(0, 4))
+        self.assert_(not Range(1, 2) <= Range(0, 4))
+        self.assert_(Range(0, 4) <= Range(1, 2))
+        self.assert_(not Range(1, 2) <= Range())
+        self.assert_(Range() <= Range(1, 2))
+        self.assert_(not Range(1) <= Range(upper=1))
+        self.assert_(Range() <= Range())
+        self.assert_(Range(empty=True) <= Range(empty=True))
+        self.assert_(Range(1, 2) <= Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(1 <= Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(not Range(1, 2) <= 1)
+
+    def test_ge_ordering(self):
+        from psycopg2.extras import Range
+        self.assert_(not Range(empty=True) >= Range(0, 4))
+        self.assert_(Range(1, 2) >= Range(0, 4))
+        self.assert_(not Range(0, 4) >= Range(1, 2))
+        self.assert_(Range(1, 2) >= Range())
+        self.assert_(not Range() >= Range(1, 2))
+        self.assert_(Range(1) >= Range(upper=1))
+        self.assert_(Range() >= Range())
+        self.assert_(Range(empty=True) >= Range(empty=True))
+        self.assert_(Range(1, 2) >= Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(not 1 >= Range(1, 2))
+        with py3_raises_typeerror():
+            self.assert_(Range(1, 2) >= 1)
+
+    def test_pickling(self):
+        from psycopg2.extras import Range
+
+        r = Range(0, 4)
+        self.assertEqual(loads(dumps(r)), r)
 
 
 def skip_if_no_range(f):
@@ -1298,8 +1532,8 @@ class RangeCasterTestCase(ConnectingTestCase):
         r = cur.fetchone()[0]
         self.assert_(isinstance(r, DateRange))
         self.assert_(not r.isempty)
-        self.assertEqual(r.lower, date(2000,1,2))
-        self.assertEqual(r.upper, date(2012,12,31))
+        self.assertEqual(r.lower, date(2000, 1, 2))
+        self.assertEqual(r.upper, date(2012, 12, 31))
         self.assert_(not r.lower_inf)
         self.assert_(not r.upper_inf)
         self.assert_(r.lower_inc)
@@ -1308,8 +1542,8 @@ class RangeCasterTestCase(ConnectingTestCase):
     def test_cast_timestamp(self):
         from psycopg2.extras import DateTimeRange
         cur = self.conn.cursor()
-        ts1 = datetime(2000,1,1)
-        ts2 = datetime(2000,12,31,23,59,59,999)
+        ts1 = datetime(2000, 1, 1)
+        ts2 = datetime(2000, 12, 31, 23, 59, 59, 999)
         cur.execute("select tsrange(%s, %s, '()')", (ts1, ts2))
         r = cur.fetchone()[0]
         self.assert_(isinstance(r, DateTimeRange))
@@ -1325,8 +1559,9 @@ class RangeCasterTestCase(ConnectingTestCase):
         from psycopg2.extras import DateTimeTZRange
         from psycopg2.tz import FixedOffsetTimezone
         cur = self.conn.cursor()
-        ts1 = datetime(2000,1,1, tzinfo=FixedOffsetTimezone(600))
-        ts2 = datetime(2000,12,31,23,59,59,999, tzinfo=FixedOffsetTimezone(600))
+        ts1 = datetime(2000, 1, 1, tzinfo=FixedOffsetTimezone(600))
+        ts2 = datetime(2000, 12, 31, 23, 59, 59, 999,
+                       tzinfo=FixedOffsetTimezone(600))
         cur.execute("select tstzrange(%s, %s, '[]')", (ts1, ts2))
         r = cur.fetchone()[0]
         self.assert_(isinstance(r, DateTimeTZRange))
@@ -1416,8 +1651,9 @@ class RangeCasterTestCase(ConnectingTestCase):
         self.assert_(isinstance(r1, DateTimeRange))
         self.assert_(r1.isempty)
 
-        ts1 = datetime(2000,1,1, tzinfo=FixedOffsetTimezone(600))
-        ts2 = datetime(2000,12,31,23,59,59,999, tzinfo=FixedOffsetTimezone(600))
+        ts1 = datetime(2000, 1, 1, tzinfo=FixedOffsetTimezone(600))
+        ts2 = datetime(2000, 12, 31, 23, 59, 59, 999,
+                       tzinfo=FixedOffsetTimezone(600))
         r = DateTimeTZRange(ts1, ts2, '(]')
         cur.execute("select %s", (r,))
         r1 = cur.fetchone()[0]
@@ -1445,7 +1681,7 @@ class RangeCasterTestCase(ConnectingTestCase):
         self.assert_(not r1.lower_inc)
         self.assert_(r1.upper_inc)
 
-        cur.execute("select %s", ([r,r,r],))
+        cur.execute("select %s", ([r, r, r],))
         rs = cur.fetchone()[0]
         self.assertEqual(len(rs), 3)
         for r1 in rs:
@@ -1454,6 +1690,9 @@ class RangeCasterTestCase(ConnectingTestCase):
             self.assert_(not r1.lower_inc)
             self.assert_(r1.upper_inc)
 
+        # clear the adapters to allow precise count by scripts/refcounter.py
+        del ext.adapters[rc.range, ext.ISQLQuote]
+
     def test_range_escaping(self):
         from psycopg2.extras import register_range
         cur = self.conn.cursor()
@@ -1466,12 +1705,12 @@ class RangeCasterTestCase(ConnectingTestCase):
                 id integer primary key,
                 range textrange)""")
 
-        bounds = [ '[)', '(]', '()', '[]' ]
-        ranges = [ TextRange(low, up, bounds[i % 4])
+        bounds = ['[)', '(]', '()', '[]']
+        ranges = [TextRange(low, up, bounds[i % 4])
             for i, (low, up) in enumerate(zip(
                 [None] + map(chr, range(1, 128)),
-                map(chr, range(1,128)) + [None],
-                ))]
+                map(chr, range(1, 128)) + [None],
+            ))]
         ranges.append(TextRange())
         ranges.append(TextRange(empty=True))
 
@@ -1505,6 +1744,9 @@ class RangeCasterTestCase(ConnectingTestCase):
             self.assertEqual(ranges[i].lower_inf, r.lower_inf)
             self.assertEqual(ranges[i].upper_inf, r.upper_inf)
 
+        # clear the adapters to allow precise count by scripts/refcounter.py
+        del ext.adapters[TextRange, ext.ISQLQuote]
+
     def test_range_not_found(self):
         from psycopg2.extras import register_range
         cur = self.conn.cursor()
@@ -1538,12 +1780,16 @@ class RangeCasterTestCase(ConnectingTestCase):
             register_range, 'rs.r1', 'FailRange', cur)
         cur.execute("rollback to savepoint x;")
 
+        # clear the adapters to allow precise count by scripts/refcounter.py
+        for r in [ra1, ra2, rars2, rars3]:
+            del ext.adapters[r.range, ext.ISQLQuote]
+
 decorate_all_tests(RangeCasterTestCase, skip_if_no_range)
 
 
 def test_suite():
     return unittest.TestLoader().loadTestsFromName(__name__)
 
+
 if __name__ == "__main__":
     unittest.main()
-
